/*  equver - solver of non-linear system of equations given in infix form
    Copyright (C) 2009  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/equver
    Contact:      equver@igor2.repo.hu
*/

%{
#include <ctype.h>
#include "infix.tab.h"
#include "equver.h"
#include "bif.h"
#include "ipath.h"

static void include(void);
%}

DIGIT	[0-9]
ID		[A-Za-z_][A-Za-z0-9_]*
PATH	[./ A-Za-z0-9_]*

%%

^include[ \t]*["]{PATH}["][\n]  { include(); }
^include[ \t]*{PATH}[\n]        { include(); }

#.*[\n]                   {}
{DIGIT}+"."{DIGIT}*       { append(yytext); yylval.num = atof(yytext); return NUM; }
0x{DIGIT}+                { append(yytext); yylval.num = atof(yytext); return NUM; }
{DIGIT}+                  { append(yytext); yylval.num = atof(yytext); return NUM; }

sin                       { append(yytext); yylval.bif = bif_sin; return BIF; }
asin                      { append(yytext); yylval.bif = bif_asin; return BIF; }
cos                       { append(yytext); yylval.bif = bif_cos; return BIF; }
acos                      { append(yytext); yylval.bif = bif_acos; return BIF; }
tan                       { append(yytext); yylval.bif = bif_tan; return BIF; }
atan                      { append(yytext); yylval.bif = bif_atan; return BIF; }
atan2                     { append(yytext); yylval.bif = bif_atan2; return BIF2; }
sqr                       { append(yytext); yylval.bif = bif_sqr; return BIF; }
sqrt                      { append(yytext); yylval.bif = bif_sqrt; return BIF; }
pow                       { append(yytext); yylval.bif = bif_pow; return BIF2; }
dist                      { append(yytext); yylval.bif = bif_dist; return BIF2; }
abs                       { append(yytext); yylval.bif = bif_abs; return BIF; }
min                       { append(yytext); yylval.bif = bif_min; return BIF2; }
max                       { append(yytext); yylval.bif = bif_max; return BIF2; }
int                       { append(yytext); yylval.bif = bif_int; return BIF; }

[:)(=\-+*/,%]             { append(yytext); return *yytext; }
[\n]                      { append(yytext); return NL; }
[ \t]                     { append(yytext); }
$verbose                  { append(yytext); return VERBOSE; }
{ID}                      { append(yytext); yylval.str = strdup(yytext); return ID; }
.                         { fprintf(stderr, "Unrecognized character in input: '%c'\n", *yytext); abort(); }

<<EOF>>                   { yypop_buffer_state(); if (!YY_CURRENT_BUFFER) yyterminate(); }

%%

static void include(void)
{
	char *s = yytext, *end;

	/* strip include and whitespace */
	while (isspace(*s)) s++;
	s += 7;
	while(isspace(*s)) s++;

	/* find end, removing any newline */
	end = s + strlen(s) - 1;
	while (((*end == '\r') || (*end == '\n')) && (end >= s)) { *end = '\0'; end--; }

	/* strip double quotes */
	if (*s == '"') {
		if (*end != *s) {
			fprintf(stderr, "Error: unterminated include file name: '%s'\n", yytext);
			abort();
		}
		s++;
		*end = '\0';
	}

	yyin = ipath_open(s);
	if (yyin == NULL) {
		fprintf(stderr, "Error: can't find %s for include\n", s);
		abort();
	}

	yypush_buffer_state(yy_create_buffer(yyin, YY_BUF_SIZE));
}
