#include <stdlib.h>
#include <stdio.h>
#include "optimize.h"

static int optimize_(code_t *eq, double *x, int eqn, vflags_t *vflags)
{
	code_t *e;
	int numvars, varid;
	double val, oldval;

	numvars = 0;
	for(e = eq; e != NULL; e = e->next)
		if (e->c == OPC_VAR) {
			numvars++;
			varid = e->data.var;
		}
	if (numvars == 1) {
		oldval = x[varid];
		x[varid] = 0;
		val = -eq_eval(eq, x);
		x[varid] = val;
		if (eq_eval(eq, x) != 0)
			x[varid] = -x[varid];
		if (eq_eval(eq, x) == 0) {
			/* remove variable if we really found the solution to this equation */
			vflags[varid] |= VF_CONSTANT;
/*			printf("OPT: equation %d is constant %f\n", eqn, val); */
			return 1;
		}
		x[varid] = oldval;
	}
	return 0;
}


void optimize_substvar(code_t **eqs, int numeqs, double *x, vflags_t *vflags)
{
	int n, varid;
	code_t *e;

	for(n = 0; n < numeqs; n++) {
		for(e = eqs[n]; e != NULL; e = e->next) {
			if ((e->c == OPC_VAR) && (vflags[e->data.var] & VF_CONSTANT)) {
				varid = e->data.var;
				e->c = OPC_NUM;
				e->data.num = x[varid];
/*				printf("Subs: %d->%f\n", varid, x[varid]);*/
			}
		}
	}
}

int optimize(code_t **eqs, int *numeqs, double *x, char **varnames, int *numvars, vflags_t *vflags, char *useful_eq, int *vleft)
{
	int n, opt, sum, left, last;
	char *del_eq;

/*	printf("\nBefore:\n");
	for(n = 0, opt = 0; n < *numeqs; n++) {
		printf("[%d]\n", n);
		print_code(eqs[n]);
	}*/

	sum = 0;
	del_eq = calloc(*numeqs, 1);
	do {
		for(n = 0, opt = 0; n < *numeqs; n++) {
			if (!del_eq[n]) {
				if (optimize_(eqs[n], x, n, vflags)) {
					opt++;
					del_eq[n] = 1;
					optimize_substvar(eqs, *numeqs, x, vflags);
				}
			}
		}
		sum += opt;
	} while(opt != 0);

/*	printf("\nAfter: vars=%d\n", *numvars);*/
	left = 0;
	for(n = 0, opt = 0; n < *numeqs; n++) {
		code_t *e;
		int v;
/*		printf("[%d] %d\n", n, del_eq[n]);
		print_code(eqs[n]);*/
		v = 0;
		for(e = eqs[n]; e != NULL; e = e->next) {
			if (e->c == OPC_VAR)
				v++;
		}
		if (v)
			left++;
		useful_eq[n] = (v != 0);
	}

	if (left == 0)
		return 0;

	/* move useless equs and constant vars at the end using swaps */
/*	printf("--before\n");
	for(n = 0; n < *numeqs; n++) {
		print_code(eqs[n]);
		printf("\n");
	}*/

	for(n = 0; n < *numeqs; n++) {
		if ((n < left) && (!useful_eq[n])) {
			for(last = (*numeqs-1); useful_eq[last] == useful_eq[n]; last--) ;
/*			printf("swap e: %d %d\n", n, last);*/
			swap_eq(n, last);
		}
	}

	*vleft = 0;
	for(n = 0; n < *numvars; n++) {
		if (!(vflags[n] & VF_CONSTANT))
			(*vleft)++;
	}


	for(n = 0; n < *numvars; n++) {
		if ((n < *vleft) && (vflags[n] & VF_CONSTANT)) {
			for(last = (*numvars-1); (vflags[last] & VF_CONSTANT) == (vflags[n] & VF_CONSTANT); last--) ;
/*			printf("swap v: %d %d\n", n, last);*/
			swap_vars(n, last);
		}
	}

/*	printf("--after\n");
	for(n = 0; n < *numeqs; n++) {
		print_code(eqs[n]);
		printf("\n");
	}*/

	return left;
}

